from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
from sklearn.metrics import roc_auc_score
from tqdm import tqdm


class Calculator:
    """Calculator class for calculating various metrics."""

    def __init__(
        self,
        df: pd.DataFrame,
        selected_columns: List[str],
        equation_type: str = "product",
        weights_for_groups: Optional[pd.Series] = None,
        equation_eval_str: Optional[str] = None,
    ) -> None:
        """Initialize Calculator.

        :param df: dataframe
        :param selected_columns: selected columns
        :param weights_for_groups: weights for group
        :param equation_eval_str: equation eval string
        """
        self.df = df
        self.df_len = len(self.df)
        self.selected_columns = selected_columns
        self.selected_values = self.df[selected_columns].values
        self.equation_type = equation_type
        self.equation_eval_str = equation_eval_str
        if weights_for_groups is None:
            self.weights_for_groups = pd.Series(
                np.ones(len(self.df)), index=self.df.index
            )
        else:
            self.weights_for_groups = weights_for_groups

    def get_overall_score(
        self,
        weights_for_equation: List[float],
    ) -> None:
        """Calculate overall score.

        :param powers_for_equation: powers for equation
        :param first_order_weights: first order weights
        """
        if len(weights_for_equation) == 2 * len(self.selected_columns):
            powers_for_equation = weights_for_equation[: len(self.selected_columns)]
            first_order_weights = weights_for_equation[len(self.selected_columns) :]
            self.df["overall_score"] = np.product(
                (1 + np.asarray(first_order_weights) * np.asarray(self.selected_values))
                ** powers_for_equation,
                axis=1,
            )
        elif self.equation_type == "product":
            self.df["overall_score"] = np.product(
                self.selected_values**weights_for_equation, axis=1
            )
        elif self.equation_type == "sum":
            weights_array = np.array(weights_for_equation).reshape(-1, 1)
            self.df["overall_score"] = self.selected_values @ weights_array

        elif self.equation_type == "free_style":
            columns = [
                self.selected_values[:, i] for i in range(self.selected_values.shape[1])
            ]
            local_dict = {"weights": weights_for_equation, "columns": columns}
            if self.equation_eval_str is not None:
                self.df["overall_score"] = eval(
                    self.equation_eval_str, globals(), local_dict
                )
            else:
                raise ValueError("equation_eval_str is not defined.")

    def calculate_wuauc(
        self,
        groupby: Optional[str],
        weights_for_equation: List,
        weights_for_groups: Optional[pd.Series] = None,
        label_column: str = "label",
        auc: bool = False,
    ) -> float:
        """Calculate weighted user AUC.

        :param groupby: groupby column
        :param weights_for_equation: weights for equation
        :param weights_for_groups: weights for group
        :param label_column: label column
        :param auc: bool, optional, default: False
        :return: AUC/WUAUC/UAUC
        """

        self.get_overall_score(weights_for_equation)
        if auc:
            result = float(
                roc_auc_score(self.df[label_column], self.df["overall_score"])
            )
        else:
            if groupby is not None:
                grouped = self.df.groupby(groupby).apply(
                    lambda x: float(roc_auc_score(x[label_column], x["overall_score"]))
                )
                if weights_for_groups is not None:
                    counts_sorted = weights_for_groups.loc[grouped.index]
                    result = float(np.average(grouped, weights=counts_sorted.values))
                else:
                    result = float(np.mean(grouped))
        return result

    def create_score_columns(
        self, boundary_dict: dict, score_column: str = "score"
    ) -> None:
        """Create score columns.

        :param boundary_dict: boundary dict
        :param score_column: score column to be converted
        """
        for k, _ in boundary_dict.items():
            self.df[f"{score_column}_lt_{k}"] = (self.df[score_column] >= k).astype(int)

    def initialize_fq_sampler(
        self,
        sample_size: int,
        score_column: str,
        slice_from: Optional[float] = None,
        slice_to: Optional[float] = None,
        log_scale: Optional[bool] = True,
        laplace_smoothing: Optional[bool] = True,
    ) -> None:
        """Initialize frequency sampler."""
        from ..sampling.frequency_sampler import FrequencySampler

        self.sampler = FrequencySampler(
            sample_size=sample_size,
            data=self.df[score_column],
            slice_from=slice_from,
            slice_to=slice_to,
            log_scale=log_scale,
            laplace_smoothing=laplace_smoothing,
        )
        self.score_column = score_column
        self.create_score_columns(
            boundary_dict=self.sampler.sample(), score_column=score_column
        )

        slice_from_condition = (
            pd.Series(True, index=self.df.index)
            if slice_from is None
            else (self.df[score_column] >= slice_from)
        )
        slice_to_condition = (
            pd.Series(True, index=self.df.index)
            if slice_to is None
            else (self.df[score_column] <= slice_to)
        )
        self.woauc_indices = self.df[slice_from_condition & slice_to_condition].index

    def calculate_woauc(
        self,
        weights_for_equation: List,
    ) -> List[float]:
        """Calculate weighted ordinal user AUC.

        :param weights_for_equation: weights for equation
        :param score_column: score column
        """
        self.get_overall_score(weights_for_equation)
        woauc = []
        for k, _ in self.sampler.boundary_dict.items():
            paritial_auc = float(
                roc_auc_score(
                    self.df.loc[self.woauc_indices][f"{self.score_column}_lt_{k}"],
                    self.df.loc[self.woauc_indices]["overall_score"],
                )
            )
            woauc.append(paritial_auc)
        return woauc

    def calculate_log_mse(self, target_column: str) -> float:
        """Calculate log mean squared error.

        :param target_column: target column
        """
        log_true = np.log(self.df[target_column] + 1)
        log_pred = np.log(self.df["overall_score"] + 1)
        mse = np.mean((log_true - log_pred) ** 2)
        return float(mse)

    def calculate_profolio_concentration(
        self, target_column: str, expected_return: Optional[float] = 0.95
    ) -> Tuple[float, float]:
        """Calculate profolio concentration.

        :param target_column: target column
        :param expected_return: expected return
        """
        df = self.df.sort_values("overall_score", ascending=False)
        sum_all = df[target_column].sum()
        df["cumulative_sum"] = df[target_column].cumsum()
        df["cumulative_ratio"] = df["cumulative_sum"] / sum_all
        threshold = df[df["cumulative_ratio"] > expected_return]["overall_score"].max()
        concentration = df[df["overall_score"] > threshold].shape[0] / self.df_len
        concentration = 1 if concentration == 0 else concentration
        return threshold, concentration

    def calculate_auc_triple_parameters(self, grid_interval: int) -> Tuple:
        """Calculate AUC triple parameters.

        :param grid_interval: grid interval
        :return: tuple of W1, W2, WUAUC
        """
        w1_values = np.linspace(0, 1, grid_interval)
        w2_values = np.linspace(0, 1, grid_interval)
        W1, W2 = np.meshgrid(w1_values, w2_values)
        WUAUC = np.zeros_like(W1)

        for i in tqdm(range(W1.shape[0]), desc="Progress"):
            for j in range(W1.shape[1]):
                w1 = W1[i, j]
                w2 = W2[i, j]
                w3 = 1 - w1 - w2
                if w3 < 0:
                    WUAUC[i, j] = np.nan
                else:
                    WUAUC[i, j] = self.calculate_wuauc(
                        groupby="user_id",
                        weights_for_equation=[w1, w2, w3],
                        weights_for_groups=self.weights_for_groups,
                    )
        return W1, W2, WUAUC



from abc import ABCMeta, abstractmethod
from typing import Optional

import optuna


class BaseObjective(metaclass=ABCMeta):
    """
    This class provides methods to optimize obejective.
    """

    def __init__(
        self,
        direction: str,
        formula: str,
        fist_order: bool = False,
        dirichlet: bool = True,
        log_file: Optional[str] = None,
    ) -> None:
        self.study = optuna.create_study(direction=direction)

    @abstractmethod
    def objective(self, trial: optuna.trial.Trial) -> float:
        """Objective function to optimize."""
        raise NotImplementedError

    def optimize(self, n_trials: int) -> None:
        """Optimize the objective."""
        self.study.optimize(self.objective, n_trials=n_trials)



import logging
from typing import List, Optional

import optuna
from optuna.trial import Trial

from ..evaluation.calculator import Calculator
from .base import BaseObjective


class MultipleObjective(BaseObjective):
    """
    This class provides methods to optimize the profolio objective.
    """

    def __init__(
        self,
        direction: str,
        weights_num: int,
        formula: str,
        power: bool = False,
        first_order: bool = False,
        first_order_lower_bound: float = 1e-3,
        first_order_upper_bound: float = 1e6,
        power_lower_bound: float = 0,
        power_upper_bound: float = 1,
        dirichlet: bool = True,
        log_file: Optional[str] = None,
    ) -> None:
        """
        Initialize with direction, weights_num, formula and dirichlet.

        Args:
            direction (str ["minimize", "maximize"]): direction to optimize.
            weights_num (int): numbers of weights to search.
            formula (str): formula of targets to calculate the objective.
            dirichlet (bool, optional): Use dirichlet distribution or not. Defaults to True.
        """
        super().__init__(direction, formula, first_order, dirichlet)
        self.calculators: List[Calculator] = []
        self.calculator_flags: List[str] = []
        self.hyperparameters: List[Optional[float]] = []
        self.target_columns: List[str] = []
        self.weights_num = weights_num
        self.formula = formula
        self.groupbys: List[Optional[str]] = []
        self.power = power
        self.power_lower_bound = power_lower_bound
        self.power_upper_bound = power_upper_bound
        self.first_order = first_order
        self.first_order_lower_bound = first_order_lower_bound
        self.first_order_upper_bound = first_order_upper_bound
        if self.power_lower_bound < 0:
            self.dirichlet = False
        else:
            self.dirichlet = dirichlet
        if log_file:
            self.build_logger(log_file)

    def build_logger(self, log_file: str) -> None:
        """Build logger for optuna.

        Args:
            log_file (str): log file path.
        """
        file_handler = logging.FileHandler(log_file)
        file_handler.setLevel(logging.INFO)
        self.logger = optuna.logging.get_logger("optuna")
        self.logger.addHandler(file_handler)

    def add_calculator(
        self,
        calculator: Calculator,
        flag: str,
        target_column: str,
        hyperparameter: Optional[float] = None,
        groupby: Optional[str] = None,
    ) -> None:
        """Add calculators to the objective.

        Args:
            calculator (Calculator): calculator building blocks.
            flag (str ["wuauc", "profolio", "logmse", ..., ect.]): type of calculator.
            target_column (str): target column to calculate.
        """
        self.calculators.append(calculator)
        self.calculator_flags.append(flag)
        self.target_columns.append(target_column)
        if hyperparameter is not None:
            self.hyperparameters.append(hyperparameter)
        else:
            self.hyperparameters.append(None)
        if groupby is not None:
            self.groupbys.append(groupby)
        else:
            self.groupbys.append(None)

    def objective(
        self,
        trial: Trial,
    ) -> float:
        """Objective function for optuna.

        Args:
            trial (Trial): optuna trial.

        Returns:
            float: objective value.
        """
        power_weights: List[float] = []
        first_order_weights: List[float] = []

        if self.power:
            if self.dirichlet:
                for i in range(self.weights_num - 1):
                    power_weights.append(
                        trial.suggest_float(
                            f"w_po_{i+1}", 0, max((1 - sum(power_weights), 0.1))
                        )
                    )
                power_weights.append(1 - sum(power_weights))
            else:
                for i in range(self.weights_num):
                    power_weights.append(
                        trial.suggest_float(
                            f"w{i+1}", self.power_lower_bound, self.power_upper_bound
                        )
                    )

        if self.first_order:
            if self.first_order_lower_bound < 0:
                log = False
            else:
                log = True
            for i in range(self.weights_num):
                first_order_weights.append(
                    trial.suggest_float(
                        f"w_fo_{i+1}",
                        self.first_order_lower_bound,
                        self.first_order_upper_bound,
                        log=log,
                    )
                )

        targets: List[float] = []
        for calculator, flag, hyperparameter, groupby, target_column in zip(
            self.calculators,
            self.calculator_flags,
            self.hyperparameters,
            self.groupbys,
            self.target_columns,
        ):
            weights: List[float] = []
            if not (self.first_order):
                weights = power_weights
            elif calculator.equation_type == "product" and self.power:
                weights = power_weights + first_order_weights
            elif calculator.equation_type == "sum":
                weights = first_order_weights
            elif calculator.equation_type == "free_style":
                weights = first_order_weights
            calculator.get_overall_score(
                weights_for_equation=weights,
            )
            if flag == "profolio":
                _, concentration = calculator.calculate_profolio_concentration(
                    target_column=target_column,
                    expected_return=hyperparameter,
                )
                targets.append(concentration)
            elif flag == "wuauc":
                wuauc = calculator.calculate_wuauc(
                    groupby=groupby,
                    label_column=target_column,
                    weights_for_equation=weights,
                )
                targets.append(wuauc)
            elif flag == "auc":
                auc = calculator.calculate_wuauc(
                    groupby=groupby,
                    label_column=target_column,
                    weights_for_equation=weights,
                    auc=True,
                )
                targets.append(auc)
            elif flag == "woauc":
                woauc = calculator.calculate_woauc(
                    weights_for_equation=weights,
                )
                targets.append(sum(woauc))
            elif flag == "logmse":
                mse = calculator.calculate_log_mse(
                    target_column=target_column,
                )
                targets.append(mse)

        local_vars = {"targets": targets, "sum": sum, "max": max, "min": min}
        result = float(eval(self.formula, {"__builtins__": None}, local_vars))
        if self.logger:
            self.logger.info(f"Trial {trial.number} finished with result: {result}")
            self.logger.info(f"targets: {targets}")
            self.logger.info(f"weights: {weights}")
            if self.first_order:
                self.logger.info(f"first_order_weights: {first_order_weights}")
        return result
